from mo_logs import logger


def mockable(enable):
    """
    FUNCTION DECORATOR TO ALLOW MOCKING THIS FUNCTION
    :param enable:
    :return:
    """
    if isinstance(enable, bool):

        def decorator(func):
            if enable:
                return Mockable(func)
            else:
                return func

        return decorator

    # ALLOW FUNCTION TO BE CALLED AS A DECORATOR
    return Mockable(enable)


class Mockable:
    def __init__(self, func):
        self.func = func

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)


def mock(mockable, *, value=None, function=None):
    """
    FUNCTION DECORATOR TO ALLOW REPLACEMENT WITH ANOTHER function (OR value)
    """
    if function is None:
        function = _constant(value)

    return Mocking(mockable, function)


class Mocking:
    """
    REPLACE mockable WITH mock FOR THE DURATION OF THIS CONTEXT
    """

    def __init__(self, mockable, mock):
        if not isinstance(mockable, Mockable):
            logger.error("expecting Mockable, not {mockable}", mockable=mockable, stack_depth=1)
        self.mockable = mockable
        self.mock = mock
        self.func = None

    def __enter__(self):
        self.func, self.mockable.func = self.mockable.func, self.mock

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.func, self.mockable.func = None, self.func

    def __call__(self, func):
        def wrapped(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return wrapped


def _constant(value):
    def f(*args, **kwargs):
        return value

    return f



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at https://www.mozilla.org/en-US/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from mo_json_config.configuration import Configuration
from mo_json_config.expander import get, get_file, expand
from mo_json_config.expand_locals import is_url

__all__ = ["get", "get_file", "expand", "configuration", "Configuration", "is_url"]


configuration = Configuration({})



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at https://www.mozilla.org/en-US/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#

import os

from mo_dots import (
    is_data,
    is_list,
    set_default,
    to_data,
    get_attr,
    listwrap,
    unwraplist,
    dict_to_data,
)
from mo_files import File
from mo_files.url import URL
from mo_future import first
from mo_logs import Except, logger, get_stacktrace

from mo_json_config.expand_locals import _replace_locals, _replace_str
from mo_json_config.schemes import scheme_loaders

CAN_NOT_READ_FILE = "Can not read file {filename}"
DEBUG = False
NOTSET = {}


def get_file(file):
    file = File(file)
    return get("file://" + file.abs_path)


LOOKBACK = 2 if DEBUG else 1


def get(url):
    """
    USE json.net CONVENTIONS TO LINK TO INLINE OTHER JSON
    """
    url = str(url)
    if "://" not in url:
        logger.error("{url} must have a prototcol (eg http://) declared", url=url)
    path = (dict_to_data({"$ref": url}), None)

    if url.startswith("file://") and url[7] != "/":
        causes = []
        candidates = [os.path.dirname(os.path.abspath(get_stacktrace(start=LOOKBACK)[0]["file"])), os.getcwd()]
        for candidate in candidates:
            if os.sep == "\\":
                base = URL("file:///" + candidate.replace(os.sep, "/").rstrip("/") + "/.")
            else:
                base = URL("file://" + candidate.rstrip("/") + "/.")
            try:
                phase1 = _replace_foreign_ref(path, base)
                break
            except Exception as cause:
                if CAN_NOT_READ_FILE in cause:
                    # lower priority cause
                    causes.append(cause)
                else:
                    causes.insert(0, cause)
        else:
            logger.error("problem replacing ref in {url}", url=url, cause=first(causes))
    else:
        phase1 = _replace_foreign_ref(path, URL(""))  # BLANK URL ONLY WORKS IF url IS ABSOLUTE

    try:
        phase2 = _replace_locals((phase1, None), url)
        return to_data(phase2)
    except Exception as cause:
        logger.error("problem replacing locals in\n{phase1}", phase1=phase1, cause=cause)


def expand(doc, doc_url="param://", params=None):
    """
    ASSUMING YOU ALREADY PULED THE doc FROM doc_url, YOU CAN STILL USE THE
    EXPANDING FEATURE

    USE mo_json_config.expand({}) TO ASSUME CURRENT WORKING DIRECTORY

    :param doc: THE DATA STRUCTURE FROM JSON SOURCE
    :param doc_url: THE URL THIS doc CAME FROM (DEFAULT USES params AS A DOCUMENT SOURCE)
    :param params: EXTRA PARAMETERS NOT FOUND IN THE doc_url PARAMETERS (WILL SUPERSEDE PARAMETERS FROM doc_url)
    :return: EXPANDED JSON-SERIALIZABLE STRUCTURE
    """
    if "://" not in doc_url:
        logger.error("{url} must have a protocol (eg https://) declared", url=doc_url)

    url = URL(doc_url)
    url.query = set_default(url.query, params)
    phase1 = _replace_foreign_ref((doc, None), url)  # BLANK URL ONLY WORKS IF url IS ABSOLUTE
    phase2 = _replace_locals((phase1, None), url)
    return to_data(phase2)


def _replace_foreign_ref(path, url):
    """
    RECURSIVELY REPLACE FOREIGN REFERENCES IN THE DATA STRUCTURE
    :param path:
    :param url:
    :return:
    """
    if url.path.endswith("/"):
        url.path = url.path[:-1]

    node = path[0]
    if is_list(node):
        output = [_replace_foreign_ref((n, path), url) for n in node]
        return output
    elif not is_data(node):
        return node

    if "$ref" not in node:
        output = {}
        for k, v in node.items():
            k = _replace_str(k, path, url)
            output[k] = _replace_foreign_ref((v, path), url)
        return output

    refs = URL(_replace_str(str(node["$ref"]), path, url))
    if "$default" in node:
        defaults = _replace_foreign_ref((node["$default"], path), url)
    else:
        defaults = NOTSET

    output = {}
    for k, v in node.items():
        if k not in ("$ref", "$default"):
            k = _replace_str(k, path, url)
            output[k] = _replace_foreign_ref((v, path), url)

    if not refs:
        if defaults is not NOTSET:
            return defaults
        return output

    ref_found = False
    ref_error = None
    ref_remain = []
    for ref in listwrap(refs):
        if not ref.scheme and not ref.path:
            # DO NOT TOUCH LOCAL REF YET
            ref_remain.append(ref)
            ref_found = True
            continue

        if not ref.scheme:
            # SCHEME RELATIVE IMPLIES SAME PROTOCOL AS LAST TIME, WHICH
            # REQUIRES THE CURRENT DOCUMENT'S SCHEME
            ref.scheme = url.scheme

        # FIND THE SCHEME AND LOAD IT
        if ref.scheme not in scheme_loaders:
            raise logger.error("unknown protocol {scheme}", scheme=ref.scheme)
        try:
            new_value = scheme_loaders[ref.scheme](ref, (node, path), url)
            ref_found = True
        except Exception as cause:
            ref_error = Except.wrap(cause)
            continue

        if ref.fragment:
            new_value = get_attr(new_value, ref.fragment)

        DEBUG and logger.note("Replace {ref} with {new_value}", ref=ref, new_value=new_value)

        if not output:
            output = new_value
        elif isinstance(output, str):
            pass  # WE HAVE A VALUE
        else:
            set_default(output, new_value)

    if not ref_found:
        if defaults is NOTSET:
            raise ref_error
    if ref_remain:
        output["$ref"] = unwraplist(ref_remain)
        output["$default"] = defaults
    if not output and defaults is not NOTSET:
        output = defaults
    DEBUG and logger.note("Return {output}", output=output)
    return output



import re

from mo_dots import is_data, is_list, set_default, from_data, is_sequence, coalesce, is_missing
from mo_files.url import URL
from mo_logs import logger

from mo_json_config.schemes import scheme_loaders

CAN_NOT_READ_FILE = "Can not read file {filename}"
DEBUG = False
NOTSET = {}


def _replace_locals(path, url):
    node = path[0]
    if is_data(node):
        for op, func in operators.items():
            if op in node:
                return func(node, path, url)
        return _replace_data(node, path, url)
    elif is_list(node):
        return [_replace_locals((n, path), url) for n in node]
    elif isinstance(node, str):
        return _replace_str(node, path[1], url)
    return node


def _replace_data(node, path, url):
    output = {}
    for k, v in node.items():
        if is_missing(v) or k in operators:
            continue
        output[k] = _replace_locals((v, path), url)
    return output


def _replace_ref(ref_node, path, url):
    ref = URL(_replace_str(str(ref_node["$ref"]), path, url))
    new_value = scheme_loaders["ref"](ref, path, url)
    defaults = _replace_default(ref_node, path, url)

    output = {}
    for k, v in ref_node.items():
        if is_missing(v) or k in ("$ref", "$default"):
            continue
        k = _replace_str(k, path, url)
        output[k] = _replace_locals((v, path), url)

    if not output:
        if defaults is NOTSET:
            return from_data(new_value)
        else:
            return from_data(set_default(new_value, defaults))
    else:
        if defaults is NOTSET:
            return from_data(set_default(output, new_value))
        else:
            return from_data(set_default(output, new_value, defaults))


def _replace_default(node, path, url):
    defaults = node.get("$default", NOTSET)
    if defaults is NOTSET:
        return NOTSET
    return _replace_locals((node["$default"], path), url)


def _replace_concat(node, path, url):
    # RECURS, DEEP COPY
    v = node.get("$concat")
    if not is_sequence(v):
        logger.error("$concat expects an array of strings")
    return coalesce(node.get("separator"), "").join(_replace_locals((vv, path), url) for vv in v)


is_url = re.compile(r"\{([0-9a-zA-Z]+://[^}]*)}")


def _replace_str(text, path, url):
    acc = []
    end = 0
    for found in is_url.finditer(text):
        acc.append(text[end: found.start()])
        try:
            ref = URL(found.group(1))
            if ref.scheme not in scheme_loaders:
                raise logger.error("unknown protocol {ref}", ref=ref)
            value = scheme_loaders[ref.scheme](ref, path, url)
            if value == None:
                raise logger.error("value not found {ref}", ref=ref)
            acc.append(value)
        except Exception as cause:
            raise logger.error("problem replacing {ref}", ref=found.group(1), cause=cause)
        end = found.end()
    if end == 0:
        return text
    return "".join(acc) + text[end:]


operators = {
    "$ref": _replace_ref,
    "$concat": _replace_concat,
    "$default": _replace_default,
}



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at https://www.mozilla.org/en-US/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#

import os

from mo_dots import get_attr
from mo_files import File
from mo_imports import delay_import
from mo_json import json2value
from mo_logs import Except, logger

from mo_json_config.convert import ini2value
from mo_future import mockable
from mo_json_config.ssm import get_ssm as _get_ssm

_replace_foreign_ref = delay_import("mo_json_config.expander._replace_foreign_ref")
_replace_locals = delay_import("mo_json_config.expander._replace_locals")

CAN_NOT_READ_FILE = "Can not read file {filename}"
DEBUG = False


@mockable
def _get_file(ref, path, url):

    if ref.path.startswith("~"):
        home_path = os.path.expanduser("~")
        if os.sep == "\\":
            home_path = "/" + home_path.replace(os.sep, "/")
        if home_path.endswith("/"):
            home_path = home_path[:-1]

        ref.path = home_path + ref.path[1::]
    elif not ref.path.startswith("/"):
        # CONVERT RELATIVE TO ABSOLUTE
        if ref.path[0] == ".":
            num_dot = 1
            while ref.path[num_dot] == ".":
                num_dot += 1

            parent = url.path.rstrip("/").split("/")[:-num_dot]
            ref.path = "/".join(parent) + ref.path[num_dot:]
        else:
            parent = url.path.rstrip("/").split("/")[:-1]
            ref.path = "/".join(parent) + "/" + ref.path

    path = ref.path if os.sep != "\\" else ref.path[1::].replace("/", "\\")

    try:
        DEBUG and logger.note("reading file {path}", path=path)
        content = File(path).read()
    except Exception as e:
        content = None
        logger.error(CAN_NOT_READ_FILE, filename=File(path).os_path, cause=e)

    try:
        new_value = json2value(content, params=ref.query, flexible=True, leaves=True)
    except Exception as e:
        e = Except.wrap(e)
        try:
            new_value = ini2value(content)
        except Exception:
            raise logger.error(CAN_NOT_READ_FILE, filename=path, cause=e)
    new_value = _replace_foreign_ref((new_value, path), ref)
    return new_value


def get_http(ref, doc_path, url):
    import requests

    params = url.query
    new_value = json2value(requests.get(str(ref)).text, params=params, flexible=True, leaves=True)
    return new_value


def _get_env(ref, doc_path, url):
    # GET ENVIRONMENT VARIABLES
    ref = ref.host
    raw_value = os.environ.get(ref)
    if not raw_value:
        logger.error("expecting environment variable with name {env_var}", env_var=ref)

    try:
        new_value = json2value(raw_value)
    except Exception as e:
        new_value = raw_value
    return new_value


def _get_keyring(ref, doc_path, url):
    try:
        import keyring
    except Exception:
        logger.error("Missing keyring: `pip install keyring` to use this feature")

    # GET PASSWORD FROM KEYRING
    service_name = ref.host
    if "@" in service_name:
        username, service_name = service_name.split("@")
    else:
        username = ref.query.username

    raw_value = keyring.get_password(service_name, username)
    if not raw_value:
        logger.error(
            "expecting password in the keyring for service_name={service_name} and username={username}",
            service_name=service_name,
            username=username,
        )

    try:
        new_value = json2value(raw_value)
    except Exception as e:
        new_value = raw_value
    return new_value


def _get_param(ref, doc_path, url):
    # GET PARAMETERS FROM url
    param = url.query
    new_value = param[ref.host]
    return new_value


def _get_value_from_fragment(ref, path, url):
    # REFER TO SELF
    frag = ref.fragment
    if frag[0] == ".":
        doc = (None, path)
        # RELATIVE
        for i, c in enumerate(frag):
            if c == ".":
                if not isinstance(doc, tuple):
                    logger.error("{frag|quote} reaches up past the root document", frag=frag)
                doc = doc[1]
            else:
                break
        new_value = get_attr(doc[0], frag[i::])
    else:
        # ABSOLUTE
        top_doc = path
        while isinstance(top_doc, tuple) and top_doc[1]:
            top_doc = top_doc[1]
        new_value = get_attr(top_doc[0], frag)
    new_value = _replace_locals((new_value, path), url)
    return new_value


def _nothing(ref, doc_path, url):
    return f"{{{ref}}}"


scheme_loaders = {
    "http": get_http,
    "https": get_http,
    "file": _get_file,
    "env": _get_env,
    "param": _get_param,
    "keyring": _get_keyring,
    "ssm": _get_ssm,
    "ref": _get_value_from_fragment,
    "scheme": _nothing,
}
